"""Checker for anything related to events"""
import discord
from .basechecker import DiscordBaseChecker
import difflib
from dpylint.assets import EVENT_LIST


class EventsChecker(DiscordBaseChecker):
    events_docs = "https://discordpy.readthedocs.io/en/latest/api.html#event-reference"

    name = "events"
    msgs = {
        "E9004": (
            "Invalid Event Name \"%s\", Did you mean %s",
            "invalid-event-name",
            "The event used must be valid (for a list of valid events see {})".format(events_docs),
        ),
        "E9005": (
            "Invalid Parameters passed to %s: %s valid parameters are %s",
            "invalid-event-params",
            "The event parameters must be valid",
        ),
    }

    def visit_asyncfunctiondef(self, event_func):
        # We find any decorators named listener, listen or event
        # These are the decorators used for listening to events
        deco = discord.utils.find(
            lambda x: x.func and (any(x.func.attrname in word for word in ("listener", "listen", "event"))),
            event_func.decorators.nodes,
        )

        # If the function does not have any of those decorators, we just return
        if not deco:
            return
        # We set the name of the event to None, we will change it once we get the name
        event_name = None

        # Check if the event has a listener
        if deco.func.attrname in ("listener", "listen"):
            # Check if the event name was passed to the listener
            if deco.args:
                # Check if the event name is valid
                if deco.args[0].value in EVENT_LIST:
                    # If it is valid then set the event name to the value passed to the decorator
                    event_name = deco.args[0].value
                else:
                    # If the event name is invalid then we add a message and also specify a possible event
                    # We specify the possible event because the user may have a typo
                    possible_event_names = difflib.get_close_matches(deco.args[0].value, EVENT_LIST, n=1, cutoff=0.1)
                    self.add_message(
                        "E9004",
                        node=deco,
                        args=(deco.args[0].value, possible_event_names[0] if possible_event_names else None),
                    )
            else:
                # If the event name was not passed to the listener, the function name is the event name.
                # We validate the event name.
                event_name = self._check_event_name(event_func)
        else:
            # If the decorator was not a listener, then it was a @bot.event decorator,
            # So the function name is the event name.
            event_name = self._check_event_name(event_func)

        if not event_name:
            # If we could not determine the event name, then we return.
            return

        # We get a list of the arguments passed to the event function.
        passed_args = [i.name for i in event_func.args.args if not i.name == "self"]

        # If the amount of arguments passed are not the same as the amount of arguments the event takes
        if len(passed_args) != EVENT_LIST[event_name]:
            # We add a message, the "(" and ")" is used to make it feel like a parameter
            # of a function rather than a parameter name
            self.add_message(
                "E9005",
                node=event_func,
                args=(
                    event_name,
                    "(" + ", ".join(passed_args) + ")",
                    "(" + ", ".join(EVENT_LIST[event_name]) + ")"
                    if event_name in EVENT_LIST
                    else None,
                ),
            )


    def _check_event_name(self, event_func):
        event_name = event_func.name

        # If the event name is not in the list of valid events, then we add a message
        if event_name not in EVENT_LIST:
            # We get the most possible event name and show it in the message
            possible_event_names = difflib.get_close_matches(event_name, EVENT_LIST, n=1, cutoff=0.1)
            self.add_message(
                "E9004",
                node=event_func,
                args=(event_name, possible_event_names[0] if possible_event_names else None),
            )
            # If the event name is invalid, then we return None
            return None
        return event_name



EVENT_LIST = {
    "on_bulk_message_delete": ["messages"],
    "on_connect": [],
    "on_disconnect": [],
    "on_error": ["event", " *args", " **kwargs"],
    "on_group_join": ["channel", " user"],
    "on_guild_available": ["guild"],
    "on_guild_channel_delete": ["channel"],
    "on_guild_channel_pins_update": ["channel", " last_pin"],
    "on_guild_channel_update": ["before", " after"],
    "on_guild_emojis_update": ["guild", " before", " after"],
    "on_guild_integrations_update": ["guild"],
    "on_guild_join": ["guild"],
    "on_guild_remove": ["guild"],
    "on_guild_role_create": ["role"],
    "on_guild_role_update": ["before", " after"],
    "on_guild_update": ["before", " after"],
    "on_invite_create": ["invite"],
    "on_invite_delete": ["invite"],
    "on_member_ban": ["guild", " user"],
    "on_member_join": ["member"],
    "on_member_unban": ["guild", " user"],
    "on_member_update": ["before", " after"],
    "on_message": ["message"],
    "on_message_delete": ["message"],
    "on_message_edit": ["before", " after"],
    "on_private_channel_delete": ["channel"],
    "on_private_channel_pins_update": ["channel", " last_pin"],
    "on_private_channel_update": ["before", " after"],
    "on_raw_bulk_message_delete": ["payload"],
    "on_raw_message_delete": ["payload"],
    "on_raw_message_edit": ["payload"],
    "on_raw_reaction_add": ["payload"],
    "on_raw_reaction_clear": ["payload"],
    "on_raw_reaction_clear_emoji": ["payload"],
    "on_raw_reaction_remove": ["payload"],
    "on_reaction_add": ["reaction", " user"],
    "on_reaction_clear": ["message", " reactions"],
    "on_reaction_clear_emoji": ["reaction"],
    "on_reaction_remove": ["reaction", " user"],
    "on_ready": [],
    "on_relationship_add": ["relationship"],
    "on_relationship_update": ["before", " after"],
    "on_resumed": [],
    "on_shard_connect": ["shard_id"],
    "on_shard_disconnect": ["shard_id"],
    "on_shard_ready": ["shard_id"],
    "on_shard_resumed": ["shard_id"],
    "on_socket_raw_receive": ["msg"],
    "on_socket_raw_send": ["payload"],
    "on_typing": ["channel", " user", " when"],
    "on_user_update": ["before", " after"],
    "on_voice_state_update": ["member", " before", " after"],
    "on_webhooks_update": ["channel"],
}



import dpylint
import pylint.testutils
import astroid


class TestHasPermissionsChecker(pylint.testutils.CheckerTestCase):
    CHECKER_CLASS = dpylint.HasPermissionsChecker

    def test_invalid_permission_value(self):
        deco_node, func_node = astroid.extract_node("""
        @commands.command()
        @commands.has_permissions(manage_messages="a") #@
        async def test(ctx): #@
            await ctx.send("hi")
        """)

        self.checker.visit_asyncfunctiondef(func_node)
        with self.assertAddsMessages(
            pylint.testutils.Message(
                msg_id='W9002',
                node=deco_node,
                args='a'
            )
        ):
            pass

    def test_invalid_permission_name(self):
        deco_node, func_node = astroid.extract_node("""
        @commands.command()
        @commands.has_permissions(manage_server=True) #@
        async def test(ctx): #@
            await ctx.send("hi")
        """)

        self.checker.visit_asyncfunctiondef(func_node)
        with self.assertAddsMessages(
            pylint.testutils.Message(
                msg_id='E9001',
                node=deco_node,
                args=('manage_server', 'manage_guild')
            )
        ):
            pass

    def test_no_permission_passed(self):
        deco_node, func_node = astroid.extract_node("""
        @commands.command()
        @commands.has_permissions() #@
        async def test(ctx): #@
            await ctx.send("hi")
        """)

        self.checker.visit_asyncfunctiondef(func_node)
        with self.assertAddsMessages(
            pylint.testutils.Message(
                msg_id='W9003',
                node=deco_node,
            )
        ):
            pass



import dpylint
import pylint.testutils
import astroid

class TestEvents(pylint.testutils.CheckerTestCase):
    CHECKER_CLASS = dpylint.EventsChecker

    def test_invalid_event_name_in_func_with_listener(self):
        func_node = astroid.extract_node("""
        @commands.Cog.listener()
        async def on_guld_jon(ctx): #@
            return ctx
        """)

        self.checker.visit_asyncfunctiondef(func_node)
        with self.assertAddsMessages(
            pylint.testutils.Message(
                msg_id='E9004',
                node=func_node,
                args=('on_guld_jon', 'on_guild_join')
            )
        ):
            pass

    def test_invalid_event_name_in_listener_with_listener(self):
        deco_node, func_node = astroid.extract_node("""
        @commands.Cog.listener("on_tpying") #@
        async def test(ctx): #@
            return ctx
        """)

        self.checker.visit_asyncfunctiondef(func_node)
        with self.assertAddsMessages(
            pylint.testutils.Message(
                msg_id='E9004',
                node=deco_node.nodes[0],
                args=('on_tpying', 'on_typing')
            )
        ):
            pass

    def test_invalid_event_params_with_listener_with_event_name_in_func(self):
        func_node = astroid.extract_node("""
        @commands.Cog.listener()
        async def on_message(message, user): #@
            return message, user
        """)

        self.checker.visit_asyncfunctiondef(func_node)
        with self.assertAddsMessages(
            pylint.testutils.Message(
                msg_id='E9005',
                node=func_node,
                args=('on_message', '(message, user)', '(message)')
            )
        ):
            pass

    def test_invalid_event_params_with_listener_with_event_name_in_listener(self):
        func_node = astroid.extract_node("""
        @commands.Cog.listener("on_ready")
        async def test(ctx):
            return ctx
        """)

        self.checker.visit_asyncfunctiondef(func_node)
        with self.assertAddsMessages(
            pylint.testutils.Message(
                msg_id='E9005',
                node=func_node,
                args=('on_ready', '(ctx)', '()')
            )
        ):
            pass



# now I know this is probably not the right way but I don't know any other way to do this
SHORTHANDS = {
    "ctx.channel.send": "ctx.send",
    "ctx.author.guild": "ctx.guild",
    "ctx.message.author": "ctx.author",
    "ctx.message.author.guild": "ctx.author",
    "ctx.message.guild": "ctx.guild",
    "ctx.message.channel":"ctx.channel",
    "ctx.message.channel.guild":"ctx.guild",
    "ctx.channel.pins":"ctx.pins"
}